@@ -33,14 +33,17 @@ struct connman_network {
 	struct connman_element element;
 	enum connman_network_type type;
 	enum connman_network_protocol protocol;
+	connman_bool_t secondary;
 	connman_bool_t available;
 	connman_bool_t connected;
 	connman_bool_t remember;
 	connman_uint8_t priority;
 	connman_uint8_t strength;
 	char *identifier;
+	char *address;
 	char *name;
 	char *node;
+	char *group;
 
 	struct connman_network_driver *driver;
 	void *driver_data;
@@ -66,6 +69,8 @@ static const char *type2string(enum connman_network_type type)
 		break;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 		return "wifi";
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		return "wimax";
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
 		return "bluetooth";
@@ -107,6 +112,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 						DBUS_TYPE_OBJECT_PATH, &path);
 	}
 
+	if (network->address != NULL)
+		connman_dbus_dict_append_variant(&dict, "Address",
+					DBUS_TYPE_STRING, &network->address);
+
 	if (network->name != NULL)
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &network->name);
@@ -426,8 +435,10 @@ static void network_destruct(struct connman_element *element)
 	g_free(network->wifi.security);
 	g_free(network->wifi.passphrase);
 
+	g_free(network->group);
 	g_free(network->node);
 	g_free(network->name);
+	g_free(network->address);
 	g_free(network->identifier);
 }
 
@@ -445,6 +456,7 @@ struct connman_network *connman_network_create(const char *identifier,
 	struct connman_network *network;
 	connman_uint8_t strength = 0;
 	const char *str;
+	char *temp;
 
 	DBG("identifier %s type %d", identifier, type);
 
@@ -456,7 +468,14 @@ struct connman_network *connman_network_create(const char *identifier,
 
 	__connman_element_initialize(&network->element);
 
-	network->element.name = g_strdup(identifier);
+	//temp = connman_dbus_encode_string(identifier);
+	temp = g_strdup(identifier);
+	if (temp == NULL) {
+		g_free(network);
+		return NULL;
+	}
+
+	network->element.name = temp;
 	network->element.type = CONNMAN_ELEMENT_TYPE_NETWORK;
 
 	network->element.network = network;
@@ -464,13 +483,12 @@ struct connman_network *connman_network_create(const char *identifier,
 
 	str = type2string(type);
 	if (str != NULL)
-		connman_element_set_static_property(&network->element,
-					"Type", DBUS_TYPE_STRING, &str);
+		connman_element_set_string(&network->element, "Type", str);
 
-	connman_element_set_static_property(&network->element,
-					"Strength", DBUS_TYPE_BYTE, &strength);
+	connman_element_set_uint8(&network->element, "Strength", strength);
 
-	network->type = type;
+	network->type       = type;
+	network->secondary  = FALSE;
 	network->identifier = g_strdup(identifier);
 
 	return network;
@@ -501,6 +519,22 @@ void connman_network_unref(struct connman_network *network)
 	connman_element_unref(&network->element);
 }
 
+const char *__connman_network_get_type(struct connman_network *network)
+{
+	return type2string(network->type);
+}
+
+/**
+ * connman_network_get_type:
+ * @network: network structure
+ *
+ * Get type of network
+ */
+enum connman_network_type connman_network_get_type(struct connman_network *network)
+{
+	return network->type;
+}
+
 /**
  * connman_network_get_identifier:
  * @network: network structure
@@ -559,6 +593,33 @@ void connman_network_set_protocol(struct connman_network *network,
 	network->protocol = protocol;
 }
 
+/**
+ * connman_network_set_group:
+ * @network: network structure
+ * @group: group name
+ *
+ * Set group name for automatic clustering
+ */
+void connman_network_set_group(struct connman_network *network,
+							const char *group)
+{
+	g_free(network->group);
+	network->group = g_strdup(group);
+}
+
+const char *__connman_network_get_group(struct connman_network *network)
+{
+	return network->group;
+}
+
+const char *__connman_network_get_ident(struct connman_network *network)
+{
+	if (network->device == NULL)
+		return NULL;
+
+	return __connman_device_get_ident(network->device);
+}
+
 /**
  * connman_network_set_available:
  * @network: network structure
@@ -619,47 +680,11 @@ connman_bool_t connman_network_get_available(struct connman_network *network)
 	return network->available;
 }
 
-/**
- * connman_network_set_connected:
- * @network: network structure
- * @connected: connected state
- *
- * Change connected state of network
- */
-int connman_network_set_connected(struct connman_network *network,
-						connman_bool_t connected)
+static gboolean set_connected(gpointer user_data)
 {
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *key = "Connected";
+	struct connman_network *network = user_data;
 
-	DBG("network %p connected %d", network, connected);
-
-	if (network->connected == connected)
-		return -EALREADY;
-
-	network->connected = connected;
-
-	if (network->registered == FALSE)
-		return 0;
-
-	signal = dbus_message_new_signal(network->element.path,
-				CONNMAN_NETWORK_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &connected);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
-
-	if (connected == TRUE) {
+	if (network->connected == TRUE) {
 		struct connman_element *element;
 		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
 
@@ -697,6 +722,53 @@ int connman_network_set_connected(struct connman_network *network,
 		__connman_device_decrease_connections(network->device);
 	}
 
+	return FALSE;
+}
+
+/**
+ * connman_network_set_connected:
+ * @network: network structure
+ * @connected: connected state
+ *
+ * Change connected state of network
+ */
+int connman_network_set_connected(struct connman_network *network,
+						connman_bool_t connected)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Connected";
+
+	DBG("network %p connected %d", network, connected);
+
+	if (network->connected == connected)
+		return -EALREADY;
+
+	network->connected = connected;
+
+	if (network->registered == FALSE) {
+		g_idle_add(set_connected, network);
+		return 0;
+	}
+
+	signal = dbus_message_new_signal(network->element.path,
+				CONNMAN_NETWORK_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return 0;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &connected);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+
+	set_connected(network);
+
 	return 0;
 }
 
@@ -789,6 +861,8 @@ int __connman_network_disconnect(struct connman_network *network)
 	if (network->connected == FALSE)
 		return -ENOTCONN;
 
+	connman_element_unregister_children(&network->element);
+
 	if (network->driver && network->driver->disconnect)
 		return network->driver->disconnect(network);
 
@@ -810,7 +884,10 @@ int connman_network_set_string(struct connman_network *network,
 {
 	DBG("network %p key %s value %s", network, key, value);
 
-	if (g_str_equal(key, "Name") == TRUE) {
+	if (g_str_equal(key, "Address") == TRUE) {
+		g_free(network->address);
+		network->address = g_strdup(value);
+	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(network->name);
 		network->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
@@ -827,7 +904,7 @@ int connman_network_set_string(struct connman_network *network,
 		network->wifi.passphrase = g_strdup(value);
 	}
 
-	return 0;
+	return connman_element_set_string(&network->element, key, value);
 }
 
 /**
@@ -842,7 +919,9 @@ const char *connman_network_get_string(struct connman_network *network,
 {
 	DBG("network %p key %s", network, key);
 
-	if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Address") == TRUE)
+		return network->address;
+	else if (g_str_equal(key, "Name") == TRUE)
 		return network->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return network->node;
@@ -853,7 +932,7 @@ const char *connman_network_get_string(struct connman_network *network,
 	else if (g_str_equal(key, "WiFi.Passphrase") == TRUE)
 		return network->wifi.passphrase;
 
-	return NULL;
+	return connman_element_get_string(&network->element, key);
 }
 
 /**
@@ -871,13 +950,10 @@ int connman_network_set_uint8(struct connman_network *network,
 
 	if (g_str_equal(key, "Priority") == TRUE)
 		network->priority = value;
-	else if (g_str_equal(key, "Strength") == TRUE) {
+	else if (g_str_equal(key, "Strength") == TRUE)
 		network->strength = value;
-		connman_element_set_static_property(&network->element,
-					"Strength", DBUS_TYPE_BYTE, &value);
-	}
 
-	return 0;
+	return connman_element_set_uint8(&network->element, key, value);
 }
 
 /**
@@ -897,7 +973,7 @@ connman_uint8_t connman_network_get_uint8(struct connman_network *network,
 	else if (g_str_equal(key, "Strength") == TRUE)
 		return network->strength;
 
-	return 0;
+	return connman_element_get_uint8(&network->element, key);
 }
 
 /**
@@ -924,7 +1000,7 @@ int connman_network_set_blob(struct connman_network *network,
 			network->wifi.ssid_len = 0;
 	}
 
-	return 0;
+	return connman_element_set_blob(&network->element, key, data, size);
 }
 
 /**
@@ -946,7 +1022,7 @@ const void *connman_network_get_blob(struct connman_network *network,
 		return network->wifi.ssid;
 	}
 
-	return NULL;
+	return connman_element_get_blob(&network->element, key, size);
 }
 
 void __connman_network_set_device(struct connman_network *network,
@@ -1034,6 +1110,22 @@ static int network_probe(struct connman_element *element)
 		return err;
 	}
 
+	network->secondary = connman_device_get_secondary(network->device);
+
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		break;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		if (network->secondary == FALSE)
+			__connman_profile_add_network(network);
+		break;
+	}
+
 	return 0;
 }
 
@@ -1049,18 +1141,60 @@ static void network_remove(struct connman_element *element)
 	if (network->driver == NULL)
 		return;
 
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		break;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		if (network->secondary == FALSE)
+			__connman_profile_remove_network(network);
+		break;
+	}
+
 	unregister_interface(element);
 
 	if (network->driver->remove)
 		network->driver->remove(network);
 }
 
+static void network_change(struct connman_element *element)
+{
+	struct connman_network *network = element->network;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (element->state != CONNMAN_ELEMENT_STATE_ERROR)
+		return;
+
+	if (element->error != CONNMAN_ELEMENT_ERROR_DHCP_FAILED)
+		return;
+
+	if (network->connected == FALSE)
+		return;
+
+	connman_element_unregister_children(element);
+
+	connman_device_set_disconnected(network->device, TRUE);
+
+	if (network->driver && network->driver->disconnect) {
+		network->driver->disconnect(network);
+		return;
+	}
+
+	network->connected = FALSE;
+}
+
 static struct connman_driver network_driver = {
 	.name		= "network",
 	.type		= CONNMAN_ELEMENT_TYPE_NETWORK,
 	.priority	= CONNMAN_DRIVER_PRIORITY_LOW,
 	.probe		= network_probe,
 	.remove		= network_remove,
+	.change		= network_change,
 };
 
 static int network_init(struct connman_device *device)
@@ -1076,6 +1210,7 @@ static int network_load(struct connman_network *network)
 	gchar *pathname, *data = NULL;
 	gsize length;
 	const char *name;
+	char *str;
 	int val;
 
 	DBG("network %p", network);
@@ -1113,13 +1248,19 @@ static int network_load(struct connman_network *network)
 	if (val > 0)
 		network->priority = val;
 
-	g_free(network->wifi.security);
-	network->wifi.security = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Security", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.security);
+		network->wifi.security = str;
+	}
 
-	g_free(network->wifi.passphrase);
-	network->wifi.passphrase = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Passphrase", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.passphrase);
+		network->wifi.passphrase = str;
+	}
 
 	g_key_file_free(keyfile);
 
@@ -1176,7 +1317,8 @@ update:
 
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
-	g_file_set_contents(pathname, data, length, NULL);
+	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
+		connman_error("Failed to store network information");
 
 done:
 	g_free(data);
